from pseudo_python.errors import PseudoPythonTypeCheckError

class Standard:
    '''
    Standard classes should respond to expand and to return
    valid nodes on expand
    '''
    pass

class StandardCall(Standard):
    '''
    converts to a standard call of the given namespace and function
    '''
    def __init__(self, namespace, function, expander=None):
        self.namespace = namespace
        self.function  = function
        self.expander = expander

    def expand(self, args):
        if not self.expander:
            q = builtin_type_check(self.namespace, self.function, None, args)[-1]
            return {'type': 'standard_call', 'namespace': self.namespace, 'function': self.function, 'args': args, 'pseudo_type': q}
        else:
            return self.expander(self.namespace, self.function, args)

class StandardMethodCall(Standard):
    '''
    converts to a method call of the same class
    '''
    def __init__(self, type, message, default=None, expander=None):
        self.type = type
        self.message = message
        self.default = default
        self.expander = expander

    def expand(self, args):
        if self.default and len(args) - 1 in self.default:
            args += self.default[len(args) - 1]
        if not self.expander:
            q = builtin_type_check(self.type, self.message, args[0], args[1:])[-1]
            return {'type': 'standard_method_call', 'receiver': args[0], 'message': self.message, 'args': args[1:], 'pseudo_type': q}
        else:
            return self.expander(self.type, self.message, args)


class StandardRegex(Standard):
    '''
    converts re.compile(r'literal') to {type: regex} node
    re.compile(variable) to re.compile(variable) standard call
    '''
    def expand(self, args):
        if args[0]['type'] == 'String':
            return {'type': 'regex', 'value': args[0]['value'], 'pseudo_type': 'Regexp'}
        else:
            return {'type': 'standard_call', 'namespace': 'regexp', 'function': 'compile', 'args': [args[0]], 'pseudo_type': 'Regexp'}


class StandardSwapper(Standard):
    def __init__(self, type, message):
        self.type = type
        self.message = message

    def expand(self, args):
        if len(args) < 2:
            raise PseudoPythonTypeCheckError('%s expects more args' % self.message)
        q = builtin_type_check(self.type, self.message, args[1], [args[0]])[-1]
        return {'type': 'standard_method_call', 'receiver': args[1], 'args': [args[0]], 'message': self.message, 'pseudo_type': q}

def to_int_expander(type, message, args):
    return len_expander(type, message, args)

def len_expander(type, message, args):
    receiver_type = args[0]['pseudo_type']
    if isinstance(receiver_type, list):
        a = receiver_type[0]
    else:
        a = receiver_type
    # print(a, message, args[0], args[1:])
    # input(0)
    if message == 'length' and 'special' in args[0]: # len(sys.argv)
        return {'type': 'standard_call', 'namespace': 'system', 'function': 'arg_count', 'args': [], 'pseudo_type': 'Int'}
    else:
        q = builtin_type_check(a, message, args[0], args[1:])
        return {'type': 'standard_method_call', 'receiver': args[0], 'message': message, 'args': [], 'pseudo_type': q[-1]}


#Some imports in Java are not useful in Ruby/Python and can be resolved immediately.
JAVA_KNOWN_IMPORTS = {
	'java.util.Map' : 'Dictionary',
	'java.util.HashMap' : 'Dictionary',
	'java.util.ArrayList' : 'List',
	'java.util.List' : 'List',
	'java.util.LinkedList' : 'List'
}


#######
#Functions#

BUILTIN_FUNCTIONS = {'System.out.println','System.out.print', 'print','println','length', 'all', 'sum','add','index','remove','get','keySet','values'}

BUILTIN_EQUIVALENT_FUNCTIONS = {'add':{'List':'push','Set':'add'},'remove':{'List':'remove','Set':'remove'},'println':'display','print':'display','size':'length','get':'index','put':'index','containsKey':'contains?','keySet':'keys','values':'values'}

BUILTIN_ARG_FUNCTIONS = {'add':'same','remove':'Int','print':'String','println':'String','size':None,'get':'Int','put':'same','containsKey':'same','keySet':None,'values':None}

BUILTIN_TYPE_FUNCTIONS = {'length':'Int','push':'Void','display':'Void','remove':'Void','index':'Void','contains?':'Boolean','add':'Void','keys':['Set','@k'],'values':['List','@v']}

PARTICULAR_FUNCTIONS = {'index'}

FUNCTION_API = {
    'global': {
        'input':    StandardCall('io', 'read'),
        'print':    StandardCall('io', 'display'),
        'str':      StandardCall('global', 'to_string'),
        'len':      StandardMethodCall('List', 'length', expander=len_expander),
        'int':      StandardMethodCall('String', 'to_int', expander=to_int_expander)
    },

    'math': {
        'log':      {
            1:      StandardCall('math', 'ln'),
            2:      StandardCall('math', 'log')
        },

        'sin':      StandardCall('math', 'sin'),
        'cos':      StandardCall('math', 'cos'),
        'tan':      StandardCall('math', 'tan'),
        'pow':      lambda left, right, pseudo_type: Node('binary_op',
                        op='**', left=left, right=right, pseudo_type=pseudo_type)
    },

    're': {
        'match':    StandardMethodCall('Regexp', 'match'),
        'sub':      StandardMethodCall('Regexp', 'replace'),
        'compile':  StandardRegex(),
        'escape':   StandardCall('regexp', 'escape')
    }
}

METHOD_API = {
    'String': {
        'split':      StandardMethodCall('String', 'split'),
        'join':       StandardSwapper('List', 'join'),
        'upper':      StandardMethodCall('String', 'upper'),
        'lower':      StandardMethodCall('String', 'lower'),
        'title':      StandardMethodCall('String', 'title'),
        'center':     StandardMethodCall('String', 'center', default={1: [{'type': 'string', 'value': ' ', 'pseudo_type': 'String'}]}),
        'index':      {
            1:        StandardMethodCall('String', 'find'),
            2:        StandardMethodCall('String', 'find_from')
        }
    },
    'List': {
        'append':   StandardMethodCall('List', 'push'),
        'pop':      StandardMethodCall('List', 'pop'),
        'insert':   {
            1:      StandardMethodCall('List', 'insert'),
            2:      StandardMethodCall('List', 'insert_at')
        },
        'remove':   StandardMethodCall('List', 'remove'),
        'extend':   StandardMethodCall('List', 'push_many'),
        'map':      StandardMethodCall('List', 'map'),
        'filter':   StandardMethodCall('List', 'filter')
    },

    'Dictionary': {
        'keys':     StandardMethodCall('Dictionary', 'keys'),
        'values':   StandardMethodCall('Dictionary', 'values'),
        '[]':       StandardMethodCall('Dictionary', 'getitem'),
        '[]=':      StandardMethodCall('Dictionary', 'setitem')
    },

    'Array': {
    },

    'Tuple': {
    },

    'Set': {
        '|':       StandardMethodCall('Set', 'union')
    },

    'Regexp': {
        'match':   StandardMethodCall('Regexp', 'match')
    },

    'RegexpMatch': {
        'group':    StandardMethodCall('RegexpMatch', 'group')
    }
}

OPERATOR_API = {
    'List':  {
        '+':    'concat',
        '*':    'repeat'
    },
    'Set':  {
        '|':    'union',
        '&':    'intersection',
        '^':    'symmetric_diff',
        '-':    'diff'
    },
    'String': {
        '+':    'concat',
        '*':    'repeat',
        '%':    'c_format'
    }
}




from pseudo_python.helpers import serialize_type

class PseudoError(Exception):
	def __init__(self, message, suggestions=None, right=None, wrong=None):
		super(PseudoError, self).__init__(message)

		self.suggestions = suggestions
		self.right = right
		self.wrong = wrong

class PseudoJavaNotTranslatableError(PseudoError):
	pass

class PseudoJavaTypeCheckError(PseudoError):
	pass

def cant_infer_error(name, line):
	return PseudoJavaTypeCheckError("pseudo-python can't infer the types for %s:\n%s\n" % (name, line),
			suggestions='you need to either:\n' +
						'	ensure %s is reachable from a call in your top scope\n' % name +
						'	e.g. a(..) in top scope, b(..) in a body, %s() in b body\n' % name +
						'	or provide type hints (see https://docs.python.org/3/library/typing.html)'
						)

def beautiful_error(exception):
	def f(function):
		def decorated(data, location=None, code=None, wrong_type=None, **options):
			return exception('%s%s%s:\n%s\n%s^' % (
				('wrong type %s\n' % serialize_type(wrong_type) if wrong_type else ''),
				data,
				(' on line %d column %d' % location) if location else '',
				code or '',
				(tab_aware(location[1], code) if location else '')),
				**options)
		return decorated
	return f

@beautiful_error(PseudoJavaTypeCheckError)
def type_check_error(data, location=None, code=None, wrong_type=None, **options):
	print(data)
	print(location)
	pass

@beautiful_error(PseudoJavaNotTranslatableError)
def translation_error(data, location=None, code=None, wrong_type=None, **options):
	pass

def tab_aware(location, code):
	'''
	if tabs in beginning of code, add tabs for them, otherwise spaces
	'''
	return ''.join(' ' if c != '\t' else '\t' for c in code[:location])



#!/usr/bin/env python3
import os
import sys
import pseudo_java
import pseudo_java.errors
import pseudo
import pseudo.errors
from colorama import init
from termcolor import colored

USAGE = '''
pseudo-java <input-filename.java> [<output-filename> / <language>] [options]

where if you omit <language>, pseudo-java will generate a 
<input-filename.pseudo.yaml> file with serialized ast 

if <output-filename> is provided, <language> will be extracted from 
the extension

it can be:
  py / python 
  rb / ruby
  js / javascript
  cs / csharp
  go

options:
	-o : Force the methods to be translated as a pure object (otherwise, they can sometimes be interpreted as functions and won't be in the same scope)

examples:
pseudo-java a.py # generates a.pseudo.yaml
pseudo-java z.py o.rb # generates a ruby translation in o.rb
'''

def main():
	if len(sys.argv) == 1:
		print(USAGE)
		return

	filename = sys.argv[1]
	with open(filename, 'r') as f:
		source = f.read()
	base, _ = os.path.splitext(filename)
	try:
		if len(sys.argv) == 2:
			clj = pseudo_java.translate_to_yaml(filename,source)
			with open('%s.pseudo.yaml' % base, 'w') as f:
				f.write(clj)
			print(colored('OK\nsaved pseudo ast as %s.pseudo.yaml' % base, 'green'))
		else:
			arg = sys.argv[2]
			if '.' in arg:
				base, language = os.path.splitext(arg)
				language = language[1:]
			else:
				language = arg
			if language not in pseudo.SUPPORTED_FORMATS:
				print(colored('%s is not supported' % language, 'red'))
				exit(1)
			if '%s.%s' % (base, pseudo.FILE_EXTENSIONS[language]) == filename:
				print(colored('this would overwrite the input file, please choose another name', 'red'))				
				exit(1)
			node = pseudo_java.translate(filename,source,sys.argv[2:])
			output = pseudo.generate(node, language)
			with open('%s.%s' % (base, pseudo.FILE_EXTENSIONS[language]), 'w') as f:
				f.write(output)	 
			print(colored('OK\nsaved as %s.%s' % (base, pseudo.FILE_EXTENSIONS[language]), 'green'))
	except pseudo_java.errors.PseudoError as e:
		print(colored(e, 'red'))
		if e.suggestions:
			print(colored(e.suggestions, 'green'))
		if e.right:
			print(colored('\nright:\n%s' % e.right, 'green'))
		if e.wrong:
			print(colored('\nwrong:\n%s' % e.wrong, 'red'))
		exit(1)
	except pseudo.errors.PseudoError as e:
		print(colored('Pseudo error:\n%s' % e, 'red'))
		exit(1)

if __name__ == '__main__':
	main()



def serialize_type(l):
    if isinstance(l, str):
        return l
    elif isinstance(l, list):
        return '%s[%s]' % (l[0], ', '.join(map(serialize_type, l[1:])))
    else:
        return str(l)

def safe_serialize_type(l):
    '''serialize only with letters, numbers and _'''

    if isinstance(l, str):
        return l
    elif isinstance(l, list):
        return '%s_%s_' % (l[0], ''.join(map(safe_serialize_type, l[1:])))
    else:
        return str(l)

def prepare_table(types, original_methods=None):
    names, args, returns = [], [], []
    max_name, max_arg, max_return = 0, 0, 0
    for name, t in types.items():
        if original_methods:
            if name in original_methods:
                name = original_methods[name]
            else:
                continue
        names.append(name)
        args.append(' '.join(serialize_type(arg) for arg in t[:-1]))
        returns.append(serialize_type(t[-1]))
        if len(name) > max_name:
            max_name = len(name)
        if len(args[-1]) > max_arg:
            max_arg = len(args[-1])
        if len(returns[-1]) > max_return:
            max_return = len(returns[-1])
    return '\n'.join(
        '  %s %s -> %s' % (name.ljust(max_name), arg_types.ljust(max_arg), return_type.ljust(max_return)) for name, arg_types, return_type in zip(names, args, returns))



class Standard:
    pass

class StandardCall(Standard):
    def __init__(self, namespace, function, expander=None):
        self.namespace = namespace
        self.function  = function
        self.expander = expander

    def expand(self, args):
        if not self.expander:
            q = builtin_type_check(self.namespace, self.function, None, args)
            return {'type': 'standard_call', 'namespace': self.namespace, 'function': self.function, 'args': args, 'pseudon_type': q[-1]}
        else:
            return self.expander(self.namespace, self.function, args)

class StandardMethodCall(Standard):
    def __init__(self, type, message, expander=None):
        self.type = type
        self.message = message
        self.expander = expander

    def expand(self, args):
        if not self.expander:
            q = builtin_type_check(self.type, self.message, args[0], args[1:])
            return {'type': 'standard_method_call', 'receiver': args[0], 'message': self.message, 'args': args[1:], 'pseudon_type': q[-1]}
        else:
            return self.expander(self.type, self.message, args)


def len_expander(type, message, args):
    receiver_type = args[0]['pseudon_type']
    if isinstance(receiver_type, tuple):
        a = receiver_type[0]
    else:
        a = receiver_type
    q = builtin_type_check(a, message, args[0], args[1:])
    return {'type': 'standard_method_call', 'receiver': args[0], 'message': message, 'args': [], 'pseudon_type': q[-1]}
    

FUNCTION_API = {
    'global': {
        'input':    StandardCall('io', 'read'),
        'print':    StandardCall('io', 'display'),
        'str':      StandardCall('global', 'to_string'),
        'len':      StandardMethodCall('List', 'length', len_expander)
    },

    'math': {
        'log':      {
            1:      StandardCall('math', 'ln'),
            2:      StandardCall('math', 'log')
        },

        'sin':      StandardCall('math', 'sin'),
        'cos':      StandardCall('math', 'cos')

    }
}

METHOD_API = {
    'List': {
        'append':   StandardMethodCall('List', 'push'),
        'pop':      StandardMethodCall('List', 'pop'),
        'insert':   {
            1:      StandardMethodCall('List', 'insert'),
            2:      StandardMethodCall('List', 'insert_at')
        },
        'remove':   StandardMethodCall('List', 'remove')
    },

    'Dictionary': {
        'keys':     StandardMethodCall('Dictionary', 'keys'),
        'values':   StandardMethodCall('Dictionary', 'values'),
        '[]':       StandardMethodCall('Dictionary', 'getitem'),
        '[]=':      StandardMethodCall('Dictionary', 'setitem')
    }
}




import pseudo_java.java_ast_translator
import yaml

def translate(name,source,options):
	j = java_ast_translator.JavaASTTranslator()
	return j.Test(name,source,options)

def translate_to_yaml(name,source):
	translation = translate(name,source,None)
	print(translation)
	noaliases = yaml.dumper.SafeDumper
	noaliases.ignore_aliases = lambda self, data: True
	return yaml.dump(translation,Dumper = noaliases)
